import sys
from types import ModuleType


def reload_package(root_module):
    """Reload package in interpreter memory.
    
    Useful during development packages and simultaneous testing them at projects.
    
    Refs: 
        https://stackoverflow.com/a/2918951/14748231
    """
    package_name = root_module.__name__

    # get a reference to each loaded module
    loaded_package_modules = dict([
        (key, value) for key, value in sys.modules.items() 
        if key.startswith(package_name) and isinstance(value, ModuleType)])

    # delete references to these loaded modules from sys.modules
    for key in loaded_package_modules:
        del sys.modules[key]

    # load each of the modules again; 
    # make old modules share state with new modules
    for key in loaded_package_modules:
        print('loading %s' % key)
        newmodule = __import__(key)
        oldmodule = loaded_package_modules[key]
        oldmodule.__dict__.clear()
        oldmodule.__dict__.update(newmodule.__dict__)



__version__ = "0.4.0b4"

from .errors.error import Error
from .errors.not_found_error import NotFoundError
from .views.view import View
from .tools.helper import Helper
from .tools.build import Build
from .emitters.emitter import Emitter
from .tools.assembler import get_root_path, get_mode
from .models.mappers.mapper import Mapper
from .models.services.service import Service
from .models.services.puft import Puft
from .models.services.database import Database, native_db
from .models.domains.cells import (
    Cell, ViewCell, ConfigCell, ServiceCell, MapperCell, EmitterCell, ErrorCell
)
from .constants.enums import (
    HTTPMethodEnum, TurboActionEnum
)
from .tools import (
    make_fail_response,
)
from .tools.decorators import (
    login_required,
)



from __future__ import annotations
import re
from copy import copy
import os
import json
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING, Callable, Type, Sequence, TypeVar
from flask import app

from flask_sqlalchemy import SQLAlchemy
from warepy import get_enum_values, log, format_message, join_paths, load_yaml

from puft.constants.enums import AppModeEnum, ConfigExtensionEnum
from puft.constants.orm_types import Model
from puft.errors.error import Error

if TYPE_CHECKING:
    # Import at type checking with future.annotations to avoid circular imports
    # and use just for typehints.
    from ..services.puft import Puft
    from ..services.database import Database
    from ...views.view import View
    from ...emitters.emitter import Emitter
    from ..mappers.mapper import Mapper
    from ..services.service import Service
    from ...constants.hints import CLIModeEnumUnion


# Set TypeVar upper bound to class defined afterwards.
# https://stackoverflow.com/a/67662276
AnyNamedCell = TypeVar("AnyNamedCell", bound="NamedCell")


@dataclass
class Cell:
    pass


@dataclass
class NamedCell(Cell):
    name: str

    @staticmethod
    def find_by_name(name: str, cells: Sequence[AnyNamedCell]) -> AnyNamedCell:
        """Traverse through given list of cells and return first one with
        specified name.
        
        Raise:
            ValueError: 
                No cell with given name found.
        """
        for cell in cells:
            if cell.name == name:
                return cell
        raise ValueError(
            format_message("No cell with given name {} found.", name))

    @staticmethod
    def map_to_name(cells: list[AnyNamedCell]) -> dict[str, AnyNamedCell]:
        """Traverse through given cells names and return dict with these cells
        as values and their names as keys."""
        cells_by_name: dict[str, AnyNamedCell] = {}
        for cell in cells:
            cells_by_name[cell.name] = cell
        return cells_by_name


@dataclass
class ConfigCell(NamedCell):
    """Config cell which can be used to load configs to appropriate instance's
    configuration by name."""
    source_by_app_mode: dict[AppModeEnum, str]

    def parse(
            self, app_mode_enum: AppModeEnum, root_path: str,
            update_with: dict[str, Any] | None = None,
            convert_keys_to_lower: bool = True) -> dict[str, Any]:
        """Parse config cell and return configuration dictionary.

        Args:
            app_mode_enum:
                App mode to run appropriate config.
            root_path:
                Path to join config cell source with.
            update_with (optional):
                Dictionary to update config cell mapping with.
                Defaults to None.
            convert_keys_to_lower (optional):
                If true, all keys from origin mapping and mapping from
                `update_with` will be converted to upper case.
        
        Raise:
            ValueError:
                If given config cell's source has unrecognized extension.
        """
        res_config: dict[str, Any] = {}

        config_by_mode: dict[AppModeEnum, dict] = self._load_config_by_mode()
        res_config = self._update_config_for_mode(config_by_mode, app_mode_enum)

        if res_config:
            self._parse_string_config_values(res_config, root_path)
        else:
            res_config = {}

        # Update given config with extra dictionary if this dictionary given
        # and not empty.
        if update_with:
            res_config.update(update_with)

        if convert_keys_to_lower:
            temp_config = {}
            for k, v in res_config.items():
                temp_config[k.lower()] = v
            res_config = temp_config

        return res_config

    def _update_config_for_mode(
            self,
            config_by_mode: dict[AppModeEnum, dict],
            app_mode_enum: AppModeEnum) -> dict:
        """Take config maps for each mode and return result config updated for
        current mode.
        
        E.g., given mode is TEST, so final config will be PROD config updated
        by DEV config and then updated by TEST config (so test keys will
        take priority).
        """ 
        prod_config = copy(config_by_mode[AppModeEnum.PROD])

        if app_mode_enum is AppModeEnum.TEST:
            dev_config = copy(config_by_mode[AppModeEnum.DEV])
            test_config = copy(config_by_mode[AppModeEnum.TEST])
            dev_config.update(test_config)
            prod_config.update(dev_config)
        elif app_mode_enum is AppModeEnum.DEV:
            dev_config = copy(config_by_mode[AppModeEnum.DEV])
            prod_config.update(dev_config)
        else:
            # Prod mode, do nothing extra.
            pass
        return prod_config
    
    def _load_config_by_mode(self) -> dict[AppModeEnum, dict]:
        config_by_mode: dict[AppModeEnum, dict] = {}
        for app_mode_enum in AppModeEnum:
            try:
                source = self.source_by_app_mode[app_mode_enum]
            except KeyError:
                # No source for such mode.
                config_by_mode[app_mode_enum] = {}
                continue
            source_extension = source[source.rfind(".")+1:]
            config_by_mode[app_mode_enum] = self._load_config_from_file(
                source_extension_enum=ConfigExtensionEnum(source_extension),
                source=source)
        return config_by_mode
    
    def _load_config_from_file(
            self,
            source_extension_enum: ConfigExtensionEnum, source: str) -> dict:
        # Fetch extension and load config from file.
        match source_extension_enum:
            case ConfigExtensionEnum.JSON:
                with open(source, "r") as config_file:
                    config = json.load(config_file)
            case ConfigExtensionEnum.YAML:
                config = load_yaml(source)
            case _:
                error_message = format_message(
                    "Unrecognized config cell source's extension.")
                raise ValueError(error_message)
        return config
    
    def _parse_string_config_values(
            self, config: dict[str, Any], root_path: str) -> None:
        for k, v in config.items():
            if type(v) == str:
                # Find environs to be requested.
                # Exclude escaped curly braces like `\{not_environ\}`.
                envs: list[str] = re.findall(r"[^\\]\{\w+\}", v)
                if envs:
                    for env in [
                                x.replace("{", "").replace("}", "")
                                for x in envs
                            ]:
                        env = env.strip()
                        real_env_value = os.getenv(env.strip())
                        if real_env_value is None:
                            raise ValueError(
                                f"Environ {env} specified in field"
                                f" {self.name}.{k} was not found")
                        else:
                            v = v.replace("{" + f"{env}" + "}", real_env_value)

                # Look for escaped curly braces and normalize them.
                v = v.replace(r"\{", "{").replace(r"\}", "}")

                # Find paths required to be joined to the root path.
                if v[0] == "." and v[1] == "/":
                    config[k] = join_paths(root_path, v)
                else:
                    config[k] = v


@dataclass
class ServiceCell(NamedCell):
    service_class: type[Service]


@dataclass
class PuftServiceCell(ServiceCell):
    """Injection cell with app itself which is required in any build."""
    service_class: type[Puft]
    mode_enum: CLIModeEnumUnion
    host: str
    port: int
    ctx_processor_func: Callable | None = None
    each_request_func: Callable | None = None
    first_request_func: Callable | None = None


@dataclass
class DatabaseServiceCell(ServiceCell):
    """Injection cell with database itself which can be applied to created application."""
    service_class: Type[Database]


@dataclass
class MapperCell(Cell):
    mapper_class: type[Mapper]
    model: type[Model]


@dataclass
class ViewCell(Cell):
    endpoint: str
    view_class: type[View]
    route: str  # Route will be the same for all methods.


@dataclass
class EmitterCell(Cell):
    emitter_class: type[Emitter]


@dataclass
class ErrorCell(Cell):
    error_class: type[Error]
    handler_function: Callable



from __future__ import annotations
from typing import TYPE_CHECKING

from warepy import log, Singleton, format_message
from flask_sqlalchemy import SQLAlchemy

from ..services.database import native_db


class Mapper(Singleton):
    """Represents database table data manipulations. Abstracted to only contain classmethods.

    Contain class variable `params` which can be assigned during program assembling individually for each children."""
    model = native_db.Model 

    @classmethod
    @log.catch
    def set_orm_model(cls, model: SQLAlchemy.Model) -> None:
        """Set mapper's orm model attribute to work with."""
        cls.model = model

    @classmethod
    @log.catch
    def filter_first(cls, **kwargs) -> SQLAlchemy.Model:
        """Filter first ORM mapped model by given kwargs and return it.
        
        Raise:
            ValueError: if there is no such ORM model in database matched given kwargs."""
        model = cls.model.query.filter_by(**kwargs).first()
        if model is None:
            error_message = format_message("There is no model {} with such parameters: {}", [cls.model, kwargs])            
            raise ValueError(error_message)
        else:
            return model

    @classmethod
    @log.catch
    def filter_all(cls, order_by: str | None = None, descending_order: bool = False, **kwargs) -> list[SQLAlchemy.Model]:
        """Filter all ORM mapped models by given kwargs and return them.

        It's possible to send argument `order_by` to order resulting instance. It should be string referencing to target object attribute.
        For example, if you want to order `Food` by `order` attribute:
        ```
            Food.filter_all(order_by="price", shop_id=12)  # Filter all food for Shop with id 12 and order them.
        ```
        Also you can append `descending_order=True` to apply descending ordering.
        
        Raise:
            ValueError: If there is no such ORM models in database matched given kwargs.
            ValuerError: If there is no such attribute to order by (wrong string given to argument `order_by`)."""
        if order_by is not None:
            # Calculate order target.
            order_target = None
            if order_by not in cls.model.__dict__:
                error_message = format_message("No such attribute in ORM model {}, matched given order argument: {}", [cls.model, order_by])
                raise ValueError(error_message)
            # Also consider asc/desc order.
            if descending_order:
                exec(f"order_target = cls.model.{order_by}.desc()")
            else:
                exec(f"order_target = cls.model.{order_by}")
            models = cls.model.query.filter_by(**kwargs).order_by(order_target).all()
        else:
            models = cls.model.query.filter_by(**kwargs).all()

        if not models:
            error_message = format_message("There is no model {} with such parameters: {}", [cls.model, kwargs])            
            raise ValueError(error_message)
        else:
            return models


import re
from functools import wraps
from typing import Callable

from warepy import log, format_message
from flask import Flask
import flask_migrate
from flask_sqlalchemy import SQLAlchemy

from .service import Service
from ...constants.enums import DatabaseTypeEnum


# Here database variable are referenced at top layer to be visible for ORMs.
# It is kinda messy, and in future it may be refactored (nothing more permanent than temporary).
native_db = SQLAlchemy()


# TODO: Fix type hinting for decorated functions under this decorator.
def migration_implemented(func: Callable):
    @wraps(func)
    def inner(self_instance, *args, **kwargs):
        if type(self_instance) is not Database:
            raise TypeError(
                format_message("Decorator migration_implemented cannot be applied to type {}.", type(self_instance))
            )
        elif not self_instance.migration:
            error_message = format_message("Migrate object hasn't been implemented yet.")
            raise AttributeError(error_message)
        else:
            return func(self_instance, *args, **kwargs)
    return inner


class Database(Service):
    """Operates over Database processes."""
    def __init__(self, config: dict) -> None:
        super().__init__(config)
        self.DEFAULT_URI = f"sqlite:///{self.config['root_path']}/sqlite3.db"

        self.native_db = native_db
        # For now service config propagated to Database domain.
        self._assign_uri_from_config(config)

    @log.catch
    def _assign_uri_from_config(self, config: dict) -> None:
        raw_uri = config.get("uri", None)  # type: str

        if not raw_uri:
            raw_uri = self.DEFAULT_URI
            log.info(f"URI for database not specified, using default")
        else:
            # Case 1: SQLite database.
            # Developer can give relative path to the database (it will be absolutized at ConfigCell.parse()),
            # by setting sqlite database extension to `.db`, e.g. `./instance/sqlite3.db`,
            # or by setting full absolute path with protocol, e.g. `sqlite:////home/user/project/instance/sqlite3.db`.
            if raw_uri.rfind(".db") != -1 or "sqlite:///" in raw_uri:
                if "sqlite:///" not in raw_uri: 
                    # Set absolute path to db.
                    # Ref: https://stackoverflow.com/a/44687471/14748231
                    self.uri = "sqlite:///" + raw_uri
                else:
                    self.uri = raw_uri
                self.type_enum = DatabaseTypeEnum.SQLITE
            # Case 2: PostgreSQL database.
            elif re.match(r"postgresql(\+\w+)?://", raw_uri):
                # No need to calculate path since psql uri should be given in full form.
                self.uri = raw_uri
                self.type_enum = DatabaseTypeEnum.PSQL
            else:
                raise ValueError(format_message("Unrecognized or yet unsupported type of database uri: {}", raw_uri))
            
            # WARNING: Never print full database uri to config, since it may contain user's password (as in case of
            # psql)
            log.info(f"Set database type: {self.type_enum.value}")

    @log.catch
    @migration_implemented
    def init_migration(self, directory: str = "migrations", multidb: bool = False) -> None:
        """Initializes migration support for the application."""
        flask_migrate.init(directory=directory, multidb=multidb)

    @log.catch
    @migration_implemented
    def migrate_migration(
        self,
        directory: str = "migrations", 
        message = None, 
        sql = False, 
        head: str = "head", 
        splice = False, 
        branch_label = None, 
        version_path = None, 
        rev_id = None
    ) -> None:
        flask_migrate.migrate(
            directory=directory,
            message=message,
            sql=sql,
            head=head,
            splice=splice,
            branch_label=branch_label,
            version_path=version_path,
            rev_id=rev_id
        )

    @log.catch
    @migration_implemented
    def upgrade_migration(
        self,
        directory: str = "migrations",
        revision: str = "head",
        sql = False,
        tag = None
    ) -> None:
        flask_migrate.upgrade(
            directory=directory,
            revision=revision,
            sql=sql,
            tag=tag
        )

    @log.catch
    def setup(self, flask_app: Flask) -> None:
        """Setup database and migration object with given Flask app."""
        flask_app.config["SQLALCHEMY_DATABASE_URI"] = self.uri
        flask_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False  # Supress warning.
        self.native_db.init_app(flask_app)

        # render_as_batch kwarg required only for sqlite3 databases to avoid ALTER TABLE issue on migrations
        # https://blog.miguelgrinberg.com/post/fixing-alter-table-errors-with-flask-migrate-and-sqlite
        if self.type_enum is DatabaseTypeEnum.SQLITE:
            is_sqlite_db = True
        else:
            is_sqlite_db = False
        self.migration = flask_migrate.Migrate(flask_app, self.native_db, render_as_batch=is_sqlite_db)

    @log.catch
    def get_native_db(self) -> SQLAlchemy:
        return self.native_db

    @log.catch
    @migration_implemented
    def get_migration(self) -> flask_migrate.Migrate:
        """Return migration object.
        
        Raise:
            AttributeError: If Migrate object hasn't implemented yet."""
        return self.migration

    @log.catch
    @migration_implemented
    def create_all_tables(self) -> None:
        self.native_db.create_all()

    @log.catch
    @migration_implemented
    def drop_all_tables(self):
        "Drop all tables."
        self.native_db.drop_all()

    @log.catch
    @migration_implemented
    def add_to_session(self, entity):
        """Place an object in the session."""
        # TODO: Add functionality to accept multiple entities as *args.
        self.native_db.session.add(entity)

    @log.catch
    @migration_implemented
    def commit_session(self):
        """Commit current transaction."""
        self.native_db.session.commit()

    @log.catch
    @migration_implemented
    def rollback_session(self):
        self.native_db.session.rollback()

    @log.catch
    @migration_implemented
    def remove_session(self):
        self.native_db.session.remove()



import os
from typing import Callable

from ..models.domains.cells import (
    ConfigCell, ServiceCell, MapperCell, ViewCell, EmitterCell, ErrorCell
)


class Build:
    """Proxy mapping class with collection of initial project instances to be builded by assembler.
    
    Should be inherited by build class in root folder."""
    def __init__(
        self,
        version: str = "",
        config_dir: str = "./src/configs",
        service_cells: list[ServiceCell] = [],
        mapper_cells: list[MapperCell] = [],
        view_cells: list[ViewCell] = [],
        emitter_cells: list[EmitterCell] = [],
        error_cells: list[ErrorCell] = [],
        shell_processors: list[Callable] = [],
        cli_cmds: list[Callable] = [],
        ctx_processor_func: Callable | None = None,
        each_request_func: Callable | None = None,
        first_request_func: Callable | None = None,
    ) -> None:
        # Use native root module path.
        # Because this class imported from root directory, root_path below will be assigned from there during class's
        # initialization.
        # NOTE: This path can be overrided by assembler in order of getting arguments on it's own initialization.
        self.root_path = os.getcwd()

        self.version = version
        self.config_dir = config_dir
        self.service_cells = service_cells
        self.mapper_cells = mapper_cells
        self.view_cells = view_cells
        self.emitter_cells = emitter_cells
        self.error_cells = error_cells
        self.shell_processors = shell_processors
        self.cli_cmds = cli_cmds
        self.ctx_processor_func = ctx_processor_func
        self.each_request_func = each_request_func
        self.first_request_func = first_request_func



from flask.views import MethodView
from warepy import log, Singleton, format_message

from ..tools.noconflict import makecls


class View(MethodView):
    """Presents API source with HTTP analog methods to be registered in app routing.
    
    Contains general methods `get`, `post`, `put` and `delete` according to same HTTP methods 
    and should be re-implemented in children classes.
    
    Refs:
        https://flask.palletsprojects.com/en/2.0.x/views/#method-views-for-apis
    """
    __metaclass__ = makecls()
    methods = ["GET", "POST", "PUT", "DELETE"]  # Methods allowed to access this view.

    # List of decorators to apply to all view's methods.
    # decorators = [log.catch]  
    # To extend decorators in child class, use `decorators = View.decorators + [your_shiny_decorator]` 
    # in your class variable definition.

    def get(self):
        error_message = format_message("Method GET is not implemented at view: {}", self.__class__.__name__)
        raise NotImplementedError(error_message)
    
    def post(self):
        error_message = format_message("Method POST is not implemented at view: {}", self.__class__.__name__)
        raise NotImplementedError(error_message)

    def put(self):
        error_message = format_message("Method PUT is not implemented at view: {}", self.__class__.__name__)
        raise NotImplementedError(error_message)

    def delete(self):
        error_message = format_message("Method DELETE is not implemented at view: {}", self.__class__.__name__)
        raise NotImplementedError(error_message)


"""Module with various tools."""
import os
import json
from typing import Any, List, Dict, Literal, Callable, Union, Tuple, Type, Sequence, TypeVar

from werkzeug.wrappers.response import Response
from flask import make_response, redirect, flash, render_template
from warepy import log, join_paths, format_message, load_yaml


@log.catch
def generate_redirect_response(url: Any, status_code: int = 302) -> Response:
    """Generate redirect response with given url and return this response."""
    response = redirect(url, code=status_code)
    return response


@log.catch
def generate_success_response(data: Any, is_json_enabled: bool, ok_status_code: int = 200, no_content_status_code: int = 204) -> Response:
    """Generate success response with given data and return this response.
    
    If `is_json_enabled` == True, dump data to json."""
    # Check if data should be dumped to json.
    if is_json_enabled:
        content = json.dumps(data)
    else:
        content = data

    response = make_response(content)

    # Set special status code if there is no content and OK otherwise.
    if content:
        response.status_code = ok_status_code
    else:
        response.status_code = no_content_status_code

    return response


@log.catch
def generate_error_response(error: Exception, status_code: int = 400) -> Response:
    """Generate error response with given error and return this response."""
    data = {"error": repr(error)}
    json_data = json.dumps(data)
    response = make_response(json_data)
    response.status_code = status_code
    return response


@log.catch
def make_fail_response(
    template_path: str, template_ctx: dict = {}, flash_message: str = "Error", flash_category: Literal["message", "error", "info", "warning"] = "error"
) -> Response:
    """Generate fail response with rendered template by given path with status code 422 compatible with turbo.js.
    
    Also able to flash message to template context.
    
    Args:
        template_path: Path to template to render.
        template_ctx (optional): Context to push to template. Defaults to empty dict.
        flash_message (optional): Message to flash to template context. Defaults to "Error".
        flash_category (optional): Category for flash message according to Flask's `flash` categories. Defaults to "error".
    """
    # Login errors should be flashed to user.
    flash(flash_message, category=flash_category)
    response = make_response(render_template(template_path, **template_ctx))
    # Always set status code of failed to form posts responses to 422 or turbo.js will raise an error.
    # Ref: https://github.com/miguelgrinberg/turbo-flask/issues/11#issuecomment-883248949
    response.status_code = 422
    return response


@log.catch
def do_or_flash(func: Callable, message: str | None = None, *args, **kwargs) -> Any:
    """Call given function with given args and kwargs and flask.flash message if the function raised an error. Return function output if there was no exception.
    
    Args:
        func: Function to execute and grab output.  
        message: Message to flash by Flask engine. Defaults to None, which means;
        - If `PUFT_MODE` is `dev`, `test`: flash error text.    
        - If `PUFT_MODE` is `prod`: flash text: "Something bad happened."  
    """
    puft_mode = os.environ["PUFT_MODE"]

    try:
        output = func(*args, **kwargs)
    except Exception as error:
        if puft_mode in ["dev", "test"] and message:
            flash(message)
        elif puft_mode in ["prod"]:
            flash("Something bad happened.")
    else:
        return output



from ..models.services.database import native_db


# Helper references for shorter writing at ORMs.
# Ignore lines added for a workaround to fix issue:
# https://github.com/microsoft/pylance-release/issues/187
Model = native_db.Model  # type: ignore
column = native_db.Column  # type: ignore
integer = native_db.Integer  # type: ignore
string = native_db.String  # type: ignore
text = native_db.Text  # type: ignore
# Longer name to avoid conflicts with Python's float object.
dbfloat = native_db.Float  # type: ignore
boolean = native_db.Boolean  # type: ignore
foreign_key = native_db.ForeignKey  # type: ignore
table = native_db.Table  # type: ignore
check_constraint = native_db.CheckConstraint  # type: ignore
relationship = native_db.relationship  # type: ignore
backref = native_db.backref  # type: ignore
pickle = native_db.PickleType  # type: ignore
binary = native_db.LargeBinary  # type: ignore
# Shorter name to avoid conflicts with Python's datetime object.
dtime = native_db.DateTime  # type: ignore



import os
import sys
import argparse
import importlib.util
from typing import get_args

from warepy import (
    format_message, join_paths, log, match_enum_containing_value,
    get_enum_values, get_union_enum_values
)

from .. import __version__ as puft_version
from .assembler import Assembler
from ..constants.hints import CLIModeEnumUnion
from ..constants.enums import (
    CLIRunEnum, CLIDatabaseEnum, CLIHelperEnum
)


def main() -> None:
    args = parse_input()

    if args.check_version:
        print(f"Puft {puft_version}")
        return
    elif args.mode[0] in get_union_enum_values(CLIModeEnumUnion):
        mode = args.mode[0] 

        # Find enum where mode assigned.
        mode_enum_class = match_enum_containing_value(
            mode, *get_args(CLIModeEnumUnion))

        # Create according enum with mode value.
        mode_enum = mode_enum_class(mode)
    else:
        raise NotImplementedError(f"Command {args.mode[0]} is not supported")
        # Mode is probably targeted to custom command.
        mode_enum = CLIHelperEnum.CUSTOM_CMD
        #!! custom_cmd = 

    # Create and build assembler.
    assembler = spawn_assembler(
        mode_enum=mode_enum,
        host=args.host,
        port=int(args.port),
        root_dir=args.root_dir,
        source_file=args.source_file,
    )
    assembler.build()

    if type(mode_enum) in get_args(CLIModeEnumUnion):
        # Call appropriate actions depend on chosen mode.
        if type(mode_enum) is CLIRunEnum:
            invoke_run(assembler)
        elif type(mode_enum) is CLIDatabaseEnum:
            with assembler.get_puft().get_native_app().app_context():
                invoke_database_change(assembler, mode_enum)
        elif type(mode_enum) is CLIHelperEnum:
            if mode_enum is CLIHelperEnum.SHELL:
                invoke_shell(assembler)
            elif mode_enum is CLIHelperEnum.CUSTOM_CMD: 
                # TODO: Custom cmds after assembler build operations.
                pass
            elif mode_enum is CLIHelperEnum.DEPLOY:
                # TODO: Implement deploy operation.
                pass

def invoke_shell(assembler: Assembler):
    """Invoke Puft interactive shell."""
    assembler.get_puft().run_shell()

def spawn_assembler(
    mode_enum: CLIModeEnumUnion, host: str, port: int, root_dir: str, source_file: str
) -> Assembler:
    """Create assembler with required parameters and return it."""
    # Add root_dir to sys.path to avoid ModuleNotFoundError during lib importing.
    sys.path.append(root_dir)

    # Load target module spec from location, where cli were called.
    module_location = os.path.join(root_dir, source_file + ".py")
    module_spec = importlib.util.spec_from_file_location(source_file, module_location)
    if module_spec and module_spec.loader:
        main_module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(main_module)
        build = main_module.build
        assembler = Assembler(
            build=build,
            mode_enum=mode_enum,
            host=host,
            port=port
        )
        return assembler
    else:
        raise NameError(format_message("Could not resolve module spec for location: {}.", module_location))


def invoke_run(
    assembler: Assembler
) -> None:
    """"""
    assembler.get_puft().run()


def invoke_database_change(
    assembler: Assembler,
    mode_enum: CLIDatabaseEnum
) -> None:
    db = assembler.get_database()

    if mode_enum is CLIDatabaseEnum.INIT:
        db.init_migration()
    elif mode_enum is CLIDatabaseEnum.MIGRATE:
        db.migrate_migration()
    elif mode_enum is CLIDatabaseEnum.UPGRADE:
        db.upgrade_migration()


def parse_input() -> argparse.Namespace:
    """Parse cli input and return argparse.Namespace object."""
    # TODO: Add help descriptions to args.
    parser = argparse.ArgumentParser()
    parser.add_argument("mode", nargs="*")
    parser.add_argument("-a", dest="host", default="127.0.0.1")
    parser.add_argument("-p", dest="port", default="5000")  # TODO: Add check if given port is integer.
    parser.add_argument("-dir", dest="root_dir", default=os.getcwd())
    parser.add_argument("-src", dest="source_file", default="build")
    parser.add_argument("-v", "--version", action="store_true", dest="check_version")
    return parser.parse_args()


if __name__ == "__main__":
    main()



from warepy import log
from puft.errors.error import Error


def handle_wildcard_error(err: Error):
    return err.expose(), err.status_code


def handle_wildcard_builtin_error(err: Exception):
    log.debug("Test!")
    return handle_wildcard_error(Error(' ; '.join(err.args), 400))



"""Module with constants and typehints."""
from typing import Any, List, Dict, Tuple, Union, Callable, Literal, get_args
from enum import Enum


class AppModeEnum(Enum):
    DEV = "dev"
    PROD = "prod"
    TEST = "test"


class CLIDatabaseEnum(Enum):
    INIT = "init"
    MIGRATE = "migrate"
    UPGRADE = "upgrade"


class CLIRunEnum(Enum):
    DEV = "dev"
    PROD = "prod"
    TEST = "test"


class CLIHelperEnum(Enum):
    SHELL = "shell"
    DEPLOY = "deploy"
    CUSTOM_CMD = "custom_cmd"


class HTTPMethodEnum(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class TurboActionEnum(Enum):
    APPEND = "append"
    PREPEND = "prepend"
    REPLACE = "replace"
    UPDATE = "update"
    REMOVE = "remove"


class DatabaseTypeEnum(Enum):
    SQLITE = "sqlite"
    PSQL = "psql"


class ConfigExtensionEnum(Enum):
    YAML = "yaml"
    JSON = "json"
